@@ -1,5 +1,6 @@
 /* Printing of backtraces and error messages
- * Copyright (C) 1996,1997,1998,1999,2000,2001, 2003, 2004, 2006, 2009, 2010, 2011 Free Software Foundation
+ * Copyright (C) 1996,1997,1998,1999,2000,2001, 2003, 2004, 2006, 2009,
+ *   2010, 2011, 2014 Free Software Foundation
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -26,9 +27,7 @@
 
 #include "libguile/_scm.h"
 
-#ifdef HAVE_UNISTD_H
 #include <unistd.h>
-#endif
 #ifdef HAVE_IO_H
 #include <io.h>
 #endif
@@ -59,32 +58,38 @@ static SCM
 boot_print_exception (SCM port, SCM frame, SCM key, SCM args)
 #define FUNC_NAME "boot-print-exception"
 {
-  scm_puts ("Throw to key ", port);
+  scm_puts_unlocked ("Throw to key ", port);
   scm_write (key, port);
-  scm_puts (" with args ", port);
+  scm_puts_unlocked (" with args ", port);
   scm_write (args, port);
   return SCM_UNSPECIFIED;
 }
 #undef FUNC_NAME
 
+static SCM print_exception_var;
+
+static void
+init_print_exception_var (void)
+{
+  print_exception_var
+    = scm_module_variable (scm_the_root_module (),
+                           scm_from_latin1_symbol ("print-exception"));
+}
+
 SCM
 scm_print_exception (SCM port, SCM frame, SCM key, SCM args)
 #define FUNC_NAME "print-exception"
 {
-  static SCM print_exception = SCM_BOOL_F;
+  static scm_i_pthread_once_t once = SCM_I_PTHREAD_ONCE_INIT;
+  scm_i_pthread_once (&once, init_print_exception_var);
 
   SCM_VALIDATE_OPOUTPORT (1, port);
   if (scm_is_true (frame))
     SCM_VALIDATE_FRAME (2, frame);
   SCM_VALIDATE_SYMBOL (3, key);
   SCM_VALIDATE_LIST (4, args);
-  
-  if (scm_is_false (print_exception))
-    print_exception =
-      scm_module_variable (scm_the_root_module (),
-                           scm_from_latin1_symbol ("print-exception"));
 
-  return scm_call_4 (scm_variable_ref (print_exception),
+  return scm_call_4 (scm_variable_ref (print_exception_var),
                      port, frame, key, args);
 }
 #undef FUNC_NAME
@@ -220,14 +225,14 @@ indent (int n, SCM port)
 {
   int i;
   for (i = 0; i < n; ++i)
-    scm_putc (' ', port);
+    scm_putc_unlocked (' ', port);
 }
 
 static void
 display_frame_expr (char *hdr, SCM exp, char *tlr, int indentation, SCM sport, SCM port, scm_print_state *pstate)
 {
   int i = 0, n;
-  scm_t_ptob_descriptor *ptob = scm_ptobs + SCM_PTOBNUM (sport);
+  scm_t_ptob_descriptor *ptob = SCM_PORT_DESCRIPTOR (sport);
   do
     {
       pstate->length = print_params[i].length;
@@ -236,7 +241,7 @@ display_frame_expr (char *hdr, SCM exp, char *tlr, int indentation, SCM sport, S
 	{
 	  pstate->level = print_params[i].level - 1;
 	  scm_iprlist (hdr, exp, tlr[0], sport, pstate);
-	  scm_puts (&tlr[1], sport);
+	  scm_puts_unlocked (&tlr[1], sport);
 	}
       else
 	{
@@ -256,14 +261,7 @@ display_frame_expr (char *hdr, SCM exp, char *tlr, int indentation, SCM sport, S
 static void
 display_application (SCM frame, int indentation, SCM sport, SCM port, scm_print_state *pstate)
 {
-  SCM proc = scm_frame_procedure (frame);
-  SCM name = (scm_is_true (scm_procedure_p (proc))
-	      ? scm_procedure_name (proc)
-	      : SCM_BOOL_F);
-  display_frame_expr ("[",
-		      scm_cons (scm_is_true (name) ? name : proc,
-				scm_frame_arguments (frame)),
-		      "]",
+  display_frame_expr ("[", scm_frame_call_representation (frame), "]",
 		      indentation,
 		      sport,
 		      port,
@@ -341,19 +339,19 @@ display_backtrace_file (frame, last_file, port, pstate)
 
   *last_file = file;
 
-  scm_puts ("In ", port);
+  scm_puts_unlocked ("In ", port);
   if (scm_is_false (file))
     if (scm_is_false (line))
-      scm_puts ("unknown file", port);
+      scm_puts_unlocked ("unknown file", port);
     else
-      scm_puts ("current input", port);
+      scm_puts_unlocked ("current input", port);
   else
     {
       pstate->writingp = 0;
       scm_iprin1 (file, port, pstate);
       pstate->writingp = 1;
     }
-  scm_puts (":\n", port);
+  scm_puts_unlocked (":\n", port);
 }
 
 static void
@@ -368,9 +366,9 @@ display_backtrace_file_and_line (SCM frame, SCM port, scm_print_state *pstate)
       if (scm_is_false (file))
 	{
 	  if (scm_is_false (line))
-	    scm_putc ('?', port);
+	    scm_putc_unlocked ('?', port);
 	  else
-	    scm_puts ("<stdin>", port);
+	    scm_puts_unlocked ("<stdin>", port);
 	}
       else
 	{
@@ -385,7 +383,7 @@ display_backtrace_file_and_line (SCM frame, SCM port, scm_print_state *pstate)
 	  pstate -> writingp = 1;
 	}
 
-      scm_putc (':', port);
+      scm_putc_unlocked (':', port);
     }
   else if (scm_is_true (line))
     {
@@ -396,10 +394,10 @@ display_backtrace_file_and_line (SCM frame, SCM port, scm_print_state *pstate)
     }
 
   if (scm_is_false (line))
-    scm_puts ("   ?", port);
+    scm_puts_unlocked ("   ?", port);
   else
     scm_intprint (scm_to_int (line) + 1, 10, port);
-  scm_puts (": ", port);
+  scm_puts_unlocked (": ", port);
 }
 
 static void
@@ -426,7 +424,7 @@ display_frame (SCM frame, int n, int nfield, int indentation,
 
   /* Display an application. */
   display_application (frame, nfield + 1 + indentation, sport, port, pstate);
-  scm_putc ('\n', port);
+  scm_putc_unlocked ('\n', port);
 }
 
 struct display_backtrace_args {
@@ -524,9 +522,9 @@ display_backtrace_body (struct display_backtrace_args *a)
 static SCM
 error_during_backtrace (void *data, SCM tag, SCM throw_args)
 {
-  SCM port = PTR2SCM (data);
+  SCM port = SCM_PACK_POINTER (data);
   
-  scm_puts ("Exception thrown while printing backtrace:\n", port);
+  scm_puts_unlocked ("Exception thrown while printing backtrace:\n", port);
   scm_print_exception (port, SCM_BOOL_F, tag, throw_args);
 
   return SCM_UNSPECIFIED;
@@ -557,7 +555,7 @@ SCM_DEFINE (scm_display_backtrace_with_highlights, "display-backtrace", 2, 3, 0,
 
   scm_internal_catch (SCM_BOOL_T,
 		      (scm_t_catch_body) display_backtrace_body, &a,
-		      (scm_t_catch_handler) error_during_backtrace, SCM2PTR (port));
+		      (scm_t_catch_handler) error_during_backtrace, SCM_UNPACK_POINTER (port));
 
   return SCM_UNSPECIFIED;
 }
@@ -587,7 +585,7 @@ SCM_DEFINE (scm_backtrace_with_highlights, "backtrace", 0, 1, 0,
     highlights = SCM_EOL;
 
   scm_newline (port);
-  scm_puts ("Backtrace:\n", port);
+  scm_puts_unlocked ("Backtrace:\n", port);
   scm_display_backtrace_with_highlights (stack, port, SCM_BOOL_F, SCM_BOOL_F,
                                          highlights);
   scm_newline (port);
